import random
from django.core.management.base import BaseCommand
from django.utils import timezone

from simulations.fitness import calculate_pathogen_score
from simulations.model_serializers import (
    serialize_individuals_and_connections,
    deserialize_individuals_and_connections,
    create_networkx_graph_from_queryset,
)
from simulations.models import Generation, Individual, Pathogen
from simulations.network_strategies import (
    generate_small_world_population,
    generate_scale_free_population,
    get_patient_zero,
    PatientZeroStrategies,
    generate_clustered_small_world_population,
)
from simulations.sampling import sample_and_mutate_pathogen_qs
from simulations.simulate import generation_tick
from simulations.skill_distribution import random_skill_distribution
from simulations.models import Experiment
from simulations.statistics import compute_graph_statistics


class Command(BaseCommand):
    def success(self, message):
        self.stdout.write(self.style.SUCCESS(message))

    def handle(self, *args, **options):
        MINIMUM_POPULATION_SIZE = 100
        ELITE_SAMPLE_SIZE = 5

        experiment_label = input("How would you like to label this experiment? ")
        population_size = int(input("Enter population size: "))

        if population_size < MINIMUM_POPULATION_SIZE:
            raise ValueError(
                "Population size must be at least {}".format(MINIMUM_POPULATION_SIZE)
            )

        number_of_time_steps = int(input("Enter number of time steps: "))
        generation_size = int(input("Enter generation size: "))
        number_of_generations = int(input("Enter number of generations: "))
        immune_response_enabled = input("Enable immune response? (y/n) ").lower() == "y"

        mitigation_factor_function = None
        population_heterogeneity = None
        if immune_response_enabled:
            mitigation_factor_function = input(
                "Enter mitigation factor function: \n1. Step-wise\n2. Linear\n"
            )
            if mitigation_factor_function == "1":
                mitigation_factor_function = (
                    Experiment.MitigationFactorFunction.STEP_WISE
                )
            elif mitigation_factor_function == "2":
                mitigation_factor_function = Experiment.MitigationFactorFunction.LINEAR
            else:
                raise ValueError("Invalid mitigation factor function")

            population_heterogeneity = input(
                "Enter population heterogeneity: \n1. Homogeneous\n2. Heterogeneous\n"
            )
            if population_heterogeneity == "1":
                population_heterogeneity = (
                    Experiment.PopulationHeterogeneity.HOMOGENEOUS
                )
            elif population_heterogeneity == "2":
                population_heterogeneity = (
                    Experiment.PopulationHeterogeneity.HETEROGENEOUS
                )
            else:
                raise ValueError("Invalid population heterogeneity")

        pathogen_immunity_duration = input(
            "Enter pathogen immunity duration: (blank -> infinite) "
        )
        if pathogen_immunity_duration == "":
            pathogen_immunity_duration = None
        else:
            try:
                pathogen_immunity_duration = int(pathogen_immunity_duration)
            except ValueError:
                raise ValueError("Invalid pathogen immunity duration")

        isolation_if_sick_enabled = input("Enable isolation if sick? (y/n) ").lower() == "y"

        patient_zero_strategy = int(
            input(
                "Enter pathogen patient zero strategy: \n1. Best-case\n2. Worst-case\n3. Random\n"
            )
        )

        if patient_zero_strategy == 1:
            PATIENT_ZERO_STRATEGY = "BEST_CASE"
        elif patient_zero_strategy == 2:
            PATIENT_ZERO_STRATEGY = "WORST_CASE"
        elif patient_zero_strategy == 3:
            PATIENT_ZERO_STRATEGY = "RANDOM"
        else:
            raise ValueError("Invalid patient zero strategy")

        if pathogen_immunity_duration == 0:
            pathogen_immunity_duration = None

        pick_network_topology = int(
            input(
                "Enter network topology: \n1. Scale Free\n2. Small World\n3. Clustered Small World\n"
            )
        )
        weighted_network = input("Weighted network? (y/n) ").lower() == "y"

        performance_score_death_weight = (input("Enter performance score death weight: (default 4.0)"))

        if performance_score_death_weight == "":
            performance_score_death_weight = 4.0
        else:
            performance_score_death_weight = float(performance_score_death_weight)

        performance_score_infection_weight = (input("Enter performance score infection weight: (default 1.0)"))

        if performance_score_infection_weight == "":
            performance_score_infection_weight = 1.0
        else:
            performance_score_infection_weight = float(performance_score_infection_weight)

        performance_score_recovery_weight = (input("Enter performance score recovery weight: default -1.0)"))

        if performance_score_recovery_weight == "":
            performance_score_recovery_weight = -1.0
        else:
            performance_score_recovery_weight = float(performance_score_recovery_weight)

        if pick_network_topology == 1:
            NETWORK_TOPOLOGY = Experiment.NetworkTopology.SCALE_FREE
            if weighted_network:
                NETWORK_TOPOLOGY = Experiment.NetworkTopology.WEIGHTED_SCALE_FREE
        elif pick_network_topology == 2:
            NETWORK_TOPOLOGY = Experiment.NetworkTopology.SMALL_WORLD
            if weighted_network:
                NETWORK_TOPOLOGY = Experiment.NetworkTopology.WEIGHTED_SMALL_WORLD
        elif pick_network_topology == 3:
            NETWORK_TOPOLOGY = Experiment.NetworkTopology.CLUSTERED_SMALL_WORLD
            if weighted_network:
                NETWORK_TOPOLOGY = (
                    Experiment.NetworkTopology.WEIGHTED_CLUSTERED_SMALL_WORLD
                )
        else:
            raise ValueError("Invalid network topology")

        DEFAULT_IMMUNE_SYSTEM = "".join(
            [
                str(random.randint(0, 1))
                for _ in range(Individual.IMMUNE_SYSTEM_STRING_LENGTH)
            ]
        )

        # First we create an experiment
        experiment = Experiment.objects.create(
            label=experiment_label,
            population_size=population_size,
            number_of_generations=number_of_generations,
            number_of_time_steps=number_of_time_steps,
            generation_size=generation_size,
            network_topology=NETWORK_TOPOLOGY,
            immune_system_base_string=DEFAULT_IMMUNE_SYSTEM,
            immunity_duration=pathogen_immunity_duration,
            immune_response_enabled=immune_response_enabled,
            mitigation_factor_function=mitigation_factor_function,
            population_heterogeneity=population_heterogeneity,
            patient_zero_strategy=PATIENT_ZERO_STRATEGY,
            isolation_if_sick_enabled=isolation_if_sick_enabled,
            performance_function_death_weight=performance_score_death_weight,
            performance_function_infection_weight=performance_score_infection_weight,
            performance_function_recovery_weight=performance_score_recovery_weight,
        )
        self.success("Experiment created")

        # Creating the first generation of populations
        self.success("Creating generations...")
        for _ in range(generation_size):
            base_pathogen_kwargs = random_skill_distribution(max_points=100)

            base_pathogen = Pathogen.objects.create(
                attack_vector="".join(
                    [
                        str(random.randint(0, 1))
                        for _ in range(Individual.IMMUNE_SYSTEM_STRING_LENGTH)
                    ]
                ),
                **base_pathogen_kwargs,
            )
            gen = Generation.objects.create(
                experiment=experiment, base_pathogen=base_pathogen
            )
            gen.save()

        self.success("Finished creating generations and base pathogens")
        self.success("Instantiating base topology...")
        self.success(f"Using {NETWORK_TOPOLOGY} topology")

        # This base population topology will be used in all runs to keep things consistent
        base_population = Individual.objects.bulk_create(
            [
                Individual(
                    # Based on setting, if enabled random string for all otherwise same string for all
                    immune_system=DEFAULT_IMMUNE_SYSTEM
                    if population_heterogeneity
                    == Experiment.PopulationHeterogeneity.HOMOGENEOUS
                    else "".join(
                        [
                            str(random.randint(0, 1))
                            for _ in range(Individual.IMMUNE_SYSTEM_STRING_LENGTH)
                        ]
                    )
                )
                for _ in range(population_size)
            ]
        )

        # This could probably be done better
        base_population_qs = Individual.objects.filter(
            id__in=[i.id for i in base_population]
        )

        if NETWORK_TOPOLOGY == Experiment.NetworkTopology.SCALE_FREE:
            generate_scale_free_population(base_population_qs, weighted=False)
        elif NETWORK_TOPOLOGY == Experiment.NetworkTopology.WEIGHTED_SCALE_FREE:
            generate_scale_free_population(base_population_qs, weighted=True)
        elif NETWORK_TOPOLOGY == Experiment.NetworkTopology.SMALL_WORLD:
            generate_small_world_population(base_population_qs, weighted=False)
        elif NETWORK_TOPOLOGY == Experiment.NetworkTopology.WEIGHTED_SMALL_WORLD:
            generate_small_world_population(base_population_qs, weighted=True)
        elif NETWORK_TOPOLOGY == Experiment.NetworkTopology.CLUSTERED_SMALL_WORLD:
            cluster_size = 50
            n_clusters = population_size // cluster_size
            generate_clustered_small_world_population(
                n_clusters,
                base_population_qs,
                cluster_size=cluster_size,
                weighted=False,
            )
        elif (
            NETWORK_TOPOLOGY
            == Experiment.NetworkTopology.WEIGHTED_CLUSTERED_SMALL_WORLD
        ):
            cluster_size = 50
            n_clusters = population_size // cluster_size
            generate_clustered_small_world_population(
                n_clusters, base_population_qs, cluster_size=cluster_size, weighted=True
            )
        else:
            raise ValueError("Invalid network topology")

        self.success("Finished instantiating base topology")
        experiment.base_individual_network_topology.set(base_population_qs)

        self.success("Aggregating network topology statistics...")
        try:
            networkx_graph = create_networkx_graph_from_queryset(base_population_qs)
            graph_stats = compute_graph_statistics(networkx_graph)
        except Exception as e:
            self.stdout.write("Error while computing graph statistics")
            self.stdout.write(e)
            self.stdout.write("Deleting experiment...")
            experiment.delete()
            raise e

        # Save the statistics in the experiment instance
        experiment.number_of_nodes = graph_stats["number_of_nodes"]
        experiment.number_of_edges = graph_stats["number_of_edges"]
        experiment.average_degree = graph_stats["average_degree"]
        experiment.average_clustering_coefficient = graph_stats[
            "average_clustering_coefficient"
        ]
        experiment.average_shortest_path_length = graph_stats[
            "average_shortest_path_length"
        ]
        experiment.diameter = graph_stats["diameter"]
        experiment.average_degree_centrality = graph_stats["average_degree_centrality"]
        experiment.average_betweenness_centrality = graph_stats[
            "average_betweenness_centrality"
        ]
        experiment.average_closeness_centrality = graph_stats[
            "average_closeness_centrality"
        ]
        experiment.average_eigenvector_centrality = graph_stats[
            "average_eigenvector_centrality"
        ]
        experiment.node_connectivity = graph_stats["node_connectivity"]
        experiment.edge_connectivity = graph_stats["edge_connectivity"]

        experiment.save()  # Don't forget to save the changes
        self.success("Finished aggregating network topology statistics")

        self.success("Starting running experiments...")

        # we copy over the base population to each generation
        for generation_count in range(1, number_of_generations + 1):
            self.success(f"Starting generation {generation_count}")
            generations = Generation.objects.filter(
                experiment=experiment, generation_number=generation_count
            )
            self.success("Serializing individuals and connections...")
            # Clone the individuals from the base population
            serialized_individuals = serialize_individuals_and_connections(
                base_population_qs
            )
            for generation in generations:
                deserialize_individuals_and_connections(
                    *serialized_individuals, generation
                )
            self.success("Finished serializing individuals and connections")

            # Main event loop for this generation
            self.success("Starting simulation loop...")

            # All of these are currently identical in all cases but allows for future flexibility
            topology_patient_zero_strategy_matrix = {
                Experiment.NetworkTopology.SCALE_FREE: {
                    "BEST_CASE": PatientZeroStrategies.MOST_CONNECTIONS,
                    "WORST_CASE": PatientZeroStrategies.LEAST_CONNECTIONS,
                    "RANDOM": PatientZeroStrategies.RANDOM,
                },
                Experiment.NetworkTopology.WEIGHTED_SCALE_FREE: {
                    "BEST_CASE": PatientZeroStrategies.MOST_CONNECTIONS,
                    "WORST_CASE": PatientZeroStrategies.LEAST_CONNECTIONS,
                    "RANDOM": PatientZeroStrategies.RANDOM,
                },
                Experiment.NetworkTopology.SMALL_WORLD: {
                    "BEST_CASE": PatientZeroStrategies.MOST_CONNECTIONS,
                    "WORST_CASE": PatientZeroStrategies.LEAST_CONNECTIONS,
                    "RANDOM": PatientZeroStrategies.RANDOM,
                },
                Experiment.NetworkTopology.WEIGHTED_SMALL_WORLD: {
                    "BEST_CASE": PatientZeroStrategies.MOST_CONNECTIONS,
                    "WORST_CASE": PatientZeroStrategies.LEAST_CONNECTIONS,
                    "RANDOM": PatientZeroStrategies.RANDOM,
                },
                Experiment.NetworkTopology.CLUSTERED_SMALL_WORLD: {
                    "BEST_CASE": PatientZeroStrategies.MOST_CONNECTIONS,
                    "WORST_CASE": PatientZeroStrategies.LEAST_CONNECTIONS,
                    "RANDOM": PatientZeroStrategies.RANDOM,
                },
                Experiment.NetworkTopology.WEIGHTED_CLUSTERED_SMALL_WORLD: {
                    "BEST_CASE": PatientZeroStrategies.MOST_CONNECTIONS,
                    "WORST_CASE": PatientZeroStrategies.LEAST_CONNECTIONS,
                    "RANDOM": PatientZeroStrategies.RANDOM,
                },
            }

            experiment.patient_zero_strategy = topology_patient_zero_strategy_matrix[
                NETWORK_TOPOLOGY
            ][PATIENT_ZERO_STRATEGY]

            for index, generation in enumerate(generations):
                self.success(f"Generation {index} of {len(generations)}")
                random_individual = get_patient_zero(
                    generation, experiment.patient_zero_strategy
                )

                pathogen_zero = generation.base_pathogen
                pathogen_zero.pk = None
                pathogen_zero.save()
                random_individual.pathogen = pathogen_zero
                random_individual.save()
                generation.number_of_infections = 1
                generation.save()
                for i in range(number_of_time_steps):
                    generation_tick(generation)

                generation.refresh_from_db()
                score = calculate_pathogen_score(generation)
                generation.pathogen_performance_score = score
                generation.save()

            generation_by_score = generations.order_by("-pathogen_performance_score")

            self.success("Preparing next generation...")
            elite_sample = generation_by_score[:ELITE_SAMPLE_SIZE]
            remaining_generation_size = generation_size - ELITE_SAMPLE_SIZE

            breeding_pool_size = remaining_generation_size // 2
            randomized_pool_size = remaining_generation_size - breeding_pool_size

            # Get breeding pool excluding the elite sample
            breeding_pool = generation_by_score[
                ELITE_SAMPLE_SIZE : ELITE_SAMPLE_SIZE + breeding_pool_size
            ]

            offspring = sample_and_mutate_pathogen_qs(
                Pathogen.objects.filter(generation__in=breeding_pool).distinct()
            )

            assert (
                breeding_pool_size + randomized_pool_size + ELITE_SAMPLE_SIZE
                == generation_size
            )

            randomized_new_pathogen_dicts = [
                random_skill_distribution(max_points=100)
                for _ in range(randomized_pool_size)
            ]

            elite_pathogens = []
            for elite in elite_sample:
                copy = elite.base_pathogen
                copy.pk = None
                copy.save()
                elite_pathogens.append(copy)

            new_pathogens = [
                Pathogen.objects.create(
                    attack_vector="".join(
                        [
                            str(random.randint(0, 1))
                            for _ in range(Individual.IMMUNE_SYSTEM_STRING_LENGTH)
                        ]
                    ),
                    **randomized_new_pathogen_dicts[i],
                )
                for i in range(randomized_pool_size)
            ]

            combined = elite_pathogens + new_pathogens + list(offspring)

            # Bulk create the new generations
            for pathogen in combined:
                Generation.objects.create(
                    experiment=experiment,
                    base_pathogen=pathogen,
                    generation_number=generation_count + 1,
                )

        self.success("Finished running experiments")
        finished_at = timezone.now()
        experiment.finished_at = finished_at
        experiment.save()
